@@ -19,7 +19,8 @@
 // CXWindowsServerTaskBarReceiver
 //
 
-CXWindowsServerTaskBarReceiver::CXWindowsServerTaskBarReceiver()
+CXWindowsServerTaskBarReceiver::CXWindowsServerTaskBarReceiver(
+				const CBufferedLogOutputter*)
 {
 	// add ourself to the task bar
 	ARCH->addReceiver(this);
@@ -53,9 +54,3 @@ CXWindowsServerTaskBarReceiver::getIcon() const
 {
 	return NULL;
 }
-
-void
-CXWindowsServerTaskBarReceiver::onStatusChanged()
-{
-	// do nothing
-}
