@@ -3,6 +3,8 @@
 // Used by synergys.rc
 //
 #define IDS_FAILED                      1
+#define IDS_INIT_FAILED                 2
+#define IDS_UNCAUGHT_EXCEPTION          3
 #define IDI_SYNERGY                     101
 #define IDI_TASKBAR_NOT_RUNNING         102
 #define IDI_TASKBAR_NOT_WORKING         103
@@ -15,13 +17,15 @@
 #define IDC_TASKBAR_QUIT                40003
 #define IDC_TASKBAR_STATUS              40004
 #define IDC_TASKBAR_LOG                 40005
+#define IDC_RELOAD_CONFIG               40006
+#define IDC_FORCE_RECONNECT             40007
 
 // Next default values for new objects
 // 
 #ifdef APSTUDIO_INVOKED
 #ifndef APSTUDIO_READONLY_SYMBOLS
 #define _APS_NEXT_RESOURCE_VALUE        109
-#define _APS_NEXT_COMMAND_VALUE         40006
+#define _APS_NEXT_COMMAND_VALUE         40008
 #define _APS_NEXT_CONTROL_VALUE         1003
 #define _APS_NEXT_SYMED_VALUE           101
 #endif
