@@ -41,6 +41,16 @@
 #	define HAVE_POSIX_SIGWAIT 1
 #endif
 
+static
+void
+setSignalSet(sigset_t* sigset)
+{
+	sigemptyset(sigset);
+	sigaddset(sigset, SIGHUP);
+	sigaddset(sigset, SIGINT);
+	sigaddset(sigset, SIGTERM);
+}
+
 //
 // CArchThreadImpl
 //
@@ -59,6 +69,7 @@ public:
 	bool				m_cancelling;
 	bool				m_exited;
 	void*				m_result;
+	void*				m_networkData;
 };
 
 CArchThreadImpl::CArchThreadImpl() :
@@ -69,7 +80,8 @@ CArchThreadImpl::CArchThreadImpl() :
 	m_cancel(false),
 	m_cancelling(false),
 	m_exited(false),
-	m_result(NULL)
+	m_result(NULL),
+	m_networkData(NULL)
 {
 	// do nothing
 }
@@ -89,6 +101,12 @@ CArchMultithreadPosix::CArchMultithreadPosix() :
 
 	s_instance = this;
 
+	// no signal handlers
+	for (size_t i = 0; i < kNUM_SIGNALS; ++i) {
+		m_signalFunc[i]     = NULL;
+		m_signalUserData[i] = NULL;
+	}
+
 	// create mutex for thread list
 	m_threadMutex = newMutex();
 
@@ -132,11 +150,36 @@ CArchMultithreadPosix::~CArchMultithreadPosix()
 	s_instance = NULL;
 }
 
+void
+CArchMultithreadPosix::setNetworkDataForCurrentThread(void* data)
+{
+	lockMutex(m_threadMutex);
+	CArchThreadImpl* thread = find(pthread_self());
+	thread->m_networkData = data;
+	unlockMutex(m_threadMutex);
+}
+
+void*
+CArchMultithreadPosix::getNetworkDataForThread(CArchThread thread)
+{
+	lockMutex(m_threadMutex);
+	void* data = thread->m_networkData;
+	unlockMutex(m_threadMutex);
+	return data;
+}
+
+CArchMultithreadPosix*
+CArchMultithreadPosix::getInstance()
+{
+	return s_instance;
+}
+
 CArchCond
 CArchMultithreadPosix::newCondVar()
 {
 	CArchCondImpl* cond = new CArchCondImpl;
 	int status = pthread_cond_init(&cond->m_cond, NULL);
+	(void)status;
 	assert(status == 0);
 	return cond;
 }
@@ -145,6 +188,7 @@ void
 CArchMultithreadPosix::closeCondVar(CArchCond cond)
 {
 	int status = pthread_cond_destroy(&cond->m_cond);
+	(void)status;
 	assert(status == 0);
 	delete cond;
 }
@@ -153,6 +197,7 @@ void
 CArchMultithreadPosix::signalCondVar(CArchCond cond)
 {
 	int status = pthread_cond_signal(&cond->m_cond);
+	(void)status;
 	assert(status == 0);
 }
 
@@ -160,6 +205,7 @@ void
 CArchMultithreadPosix::broadcastCondVar(CArchCond cond)
 {
 	int status = pthread_cond_broadcast(&cond->m_cond);
+	(void)status;
 	assert(status == 0);
 }
 
@@ -167,73 +213,43 @@ bool
 CArchMultithreadPosix::waitCondVar(CArchCond cond,
 							CArchMutex mutex, double timeout)
 {
+	// we can't wait on a condition variable and also wake it up for
+	// cancellation since we don't use posix cancellation.  so we
+	// must wake up periodically to check for cancellation.  we
+	// can't simply go back to waiting after the check since the
+	// condition may have changed and we'll have lost the signal.
+	// so we have to return to the caller.  since the caller will
+	// always check for spurious wakeups the only drawback here is
+	// performance:  we're waking up a lot more than desired.
+	static const double maxCancellationLatency = 0.1;
+	if (timeout < 0.0 || timeout > maxCancellationLatency) {
+		timeout = maxCancellationLatency;
+	}
+
+	// see if we should cancel this thread
+	testCancelThread();
+
 	// get final time
 	struct timeval now;
 	gettimeofday(&now, NULL);
 	struct timespec finalTime;
-	finalTime.tv_sec  = now.tv_sec;
-	finalTime.tv_nsec = now.tv_usec * 1000;
-	if (timeout >= 0.0) {
-		const long timeout_sec  = (long)timeout;
-		const long timeout_nsec = (long)(1.0e+9 * (timeout - timeout_sec));
-		finalTime.tv_sec  += timeout_sec;
-		finalTime.tv_nsec += timeout_nsec;
-		if (finalTime.tv_nsec >= 1000000000) {
-			finalTime.tv_nsec -= 1000000000;
-			finalTime.tv_sec  += 1;
-		}
+	finalTime.tv_sec   = now.tv_sec;
+	finalTime.tv_nsec  = now.tv_usec * 1000;
+	long timeout_sec   = (long)timeout;
+	long timeout_nsec  = (long)(1.0e+9 * (timeout - timeout_sec));
+	finalTime.tv_sec  += timeout_sec;
+	finalTime.tv_nsec += timeout_nsec;
+	if (finalTime.tv_nsec >= 1000000000) {
+		finalTime.tv_nsec -= 1000000000;
+		finalTime.tv_sec  += 1;
 	}
 
-	// repeat until we reach the final time
-	int status;
-	for (;;) {
-		// get current time
-		gettimeofday(&now, NULL);
-		struct timespec endTime;
-		endTime.tv_sec  = now.tv_sec;
-		endTime.tv_nsec = now.tv_usec * 1000;
+	// wait
+	int status = pthread_cond_timedwait(&cond->m_cond,
+							&mutex->m_mutex, &finalTime);
 
-		// done if past final timeout
-		if (timeout >= 0.0) {
-			if (endTime.tv_sec > finalTime.tv_sec ||
-				(endTime.tv_sec  == finalTime.tv_sec &&
-				 endTime.tv_nsec >= finalTime.tv_nsec)) {
-				status = ETIMEDOUT;
-				break;
-			}
-		}
-
-		// compute the next timeout
-		endTime.tv_nsec += 50000000;
-		if (endTime.tv_nsec >= 1000000000) {
-			endTime.tv_nsec -= 1000000000;
-			endTime.tv_sec  += 1;
-		}
-
-		// don't wait past final timeout
-		if (timeout >= 0.0) {
-			if (endTime.tv_sec > finalTime.tv_sec ||
-				(endTime.tv_sec  == finalTime.tv_sec &&
-				 endTime.tv_nsec >= finalTime.tv_nsec)) {
-				endTime = finalTime;
-			}
-		}
-
-		// see if we should cancel this thread
-		testCancelThread();
-
-		// wait
-		status = pthread_cond_timedwait(&cond->m_cond,
-							&mutex->m_mutex, &endTime);
-
-		// check for cancel again
-		testCancelThread();
-
-		// check wait status
-		if (status != ETIMEDOUT && status != EINTR) {
-			break;
-		}
-	}
+	// check for cancel again
+	testCancelThread();
 
 	switch (status) {
 	case 0:
@@ -252,14 +268,12 @@ CArchMultithreadPosix::waitCondVar(CArchCond cond,
 CArchMutex
 CArchMultithreadPosix::newMutex()
 {
+	pthread_mutexattr_t attr;
+	int status = pthread_mutexattr_init(&attr);
+	assert(status == 0);
 	CArchMutexImpl* mutex = new CArchMutexImpl;
-	int status = pthread_mutex_init(&mutex->m_mutex, NULL);
+	status = pthread_mutex_init(&mutex->m_mutex, &attr);
 	assert(status == 0);
-/*
-	status = pthread_mutexattr_settype(&mutex->m_mutex,
-							PTHREAD_MUTEX_RECURSIVE);
-	assert(status == 0);
-*/
 	return mutex;
 }
 
@@ -267,6 +281,7 @@ void
 CArchMultithreadPosix::closeMutex(CArchMutex mutex)
 {
 	int status = pthread_mutex_destroy(&mutex->m_mutex);
+	(void)status;
 	assert(status == 0);
 	delete mutex;
 }
@@ -341,13 +356,6 @@ CArchMultithreadPosix::newThread(ThreadFunc func, void* data)
 	thread->m_func          = func;
 	thread->m_userData      = data;
 
-	// mask some signals in all threads except the main thread
-	sigset_t sigset, oldsigset;
-	sigemptyset(&sigset);
-	sigaddset(&sigset, SIGINT);
-	sigaddset(&sigset, SIGTERM);
-	pthread_sigmask(SIG_BLOCK, &sigset, &oldsigset);
-
 	// create the thread.  pthread_create() on RedHat 7.2 smp fails
 	// if passed a NULL attr so use a default attr.
 	pthread_attr_t attr;
@@ -358,9 +366,6 @@ CArchMultithreadPosix::newThread(ThreadFunc func, void* data)
 		pthread_attr_destroy(&attr);
 	}
 
-	// restore signals
-	pthread_sigmask(SIG_SETMASK, &oldsigset, NULL);
-
 	// check if thread was started
 	if (status != 0) {
 		// failed to start thread so clean up
@@ -517,13 +522,6 @@ CArchMultithreadPosix::wait(CArchThread target, double timeout)
 	}
 }
 
-IArchMultithread::EWaitResult
-CArchMultithreadPosix::waitForEvent(CArchThread, double /*timeout*/)
-{
-	// not implemented
-	return kTimeout;
-}
-
 bool
 CArchMultithreadPosix::isSameThread(CArchThread thread1, CArchThread thread2)
 {
@@ -554,16 +552,38 @@ CArchMultithreadPosix::getIDOfThread(CArchThread thread)
 	return thread->m_id;
 }
 
+void
+CArchMultithreadPosix::setSignalHandler(
+				ESignal signal, SignalFunc func, void* userData)
+{
+	lockMutex(m_threadMutex);
+	m_signalFunc[signal]     = func;
+	m_signalUserData[signal] = userData;
+	unlockMutex(m_threadMutex);
+}
+
+void
+CArchMultithreadPosix::raiseSignal(ESignal signal) 
+{
+	lockMutex(m_threadMutex);
+	if (m_signalFunc[signal] != NULL) {
+		m_signalFunc[signal](signal, m_signalUserData[signal]);
+		pthread_kill(m_mainThread->m_thread, SIGWAKEUP);
+	}
+	else if (signal == kINTERRUPT || signal == kTERMINATE) {
+		ARCH->cancelThread(m_mainThread);
+	}
+	unlockMutex(m_threadMutex);
+}
+
 void
 CArchMultithreadPosix::startSignalHandler()
 {
 	// set signal mask.  the main thread blocks these signals and
 	// the signal handler thread will listen for them.
-	sigset_t sigset;
-	sigemptyset(&sigset);
-	sigaddset(&sigset, SIGINT);
-	sigaddset(&sigset, SIGTERM);
-	pthread_sigmask(SIG_BLOCK, &sigset, NULL);
+	sigset_t sigset, oldsigset;
+	setSignalSet(&sigset);
+	pthread_sigmask(SIG_BLOCK, &sigset, &oldsigset);
 
 	// fire up the INT and TERM signal handler thread.  we could
 	// instead arrange to catch and handle these signals but
@@ -574,16 +594,13 @@ CArchMultithreadPosix::startSignalHandler()
 	if (status == 0) {
 		status = pthread_create(&m_signalThread, &attr,
 							&CArchMultithreadPosix::threadSignalHandler,
-							m_mainThread);
+							NULL);
 		pthread_attr_destroy(&attr);
 	}
 	if (status != 0) {
 		// can't create thread to wait for signal so don't block
 		// the signals.
-		sigemptyset(&sigset);
-		sigaddset(&sigset, SIGINT);
-		sigaddset(&sigset, SIGTERM);
-		pthread_sigmask(SIG_UNBLOCK, &sigset, NULL);
+		pthread_sigmask(SIG_UNBLOCK, &oldsigset, NULL);
 	}
 }
 
@@ -731,18 +748,14 @@ CArchMultithreadPosix::threadCancel(int)
 }
 
 void*
-CArchMultithreadPosix::threadSignalHandler(void* vrep)
+CArchMultithreadPosix::threadSignalHandler(void*)
 {
-	CArchThreadImpl* mainThread = reinterpret_cast<CArchThreadImpl*>(vrep);
-
 	// detach
 	pthread_detach(pthread_self());
 
 	// add signal to mask
 	sigset_t sigset;
-	sigemptyset(&sigset);
-	sigaddset(&sigset, SIGINT);
-	sigaddset(&sigset, SIGTERM);
+	setSignalSet(&sigset);
 
 	// also wait on SIGABRT.  on linux (others?) this thread (process)
 	// will persist after all the other threads evaporate due to an
@@ -764,8 +777,25 @@ CArchMultithreadPosix::threadSignalHandler(void* vrep)
 		sigwait(&sigset);
 #endif
 
-		// if we get here then the signal was raised.  cancel the main
-		// thread so it can shut down cleanly.
-		ARCH->cancelThread(mainThread);
+		// if we get here then the signal was raised
+		switch (signal) {
+		case SIGINT:
+			ARCH->raiseSignal(kINTERRUPT);
+			break;
+
+		case SIGTERM:
+			ARCH->raiseSignal(kTERMINATE);
+			break;
+
+		case SIGHUP:
+			ARCH->raiseSignal(kHANGUP);
+			break;
+
+		default:
+			// ignore
+			break;
+		}
 	}
+
+	return NULL;
 }
