@@ -17,10 +17,7 @@
 
 #include "IInterface.h"
 
-class CInputStreamFilter;
-class COutputStreamFilter;
-class IInputStream;
-class IOutputStream;
+class IStream;
 
 //! Stream filter factory interface
 /*!
@@ -28,21 +25,12 @@ This interface provides factory methods to create stream filters.
 */
 class IStreamFilterFactory : public IInterface {
 public:
-	//! Create input filter
+	//! Create filter
 	/*!
-	Create and return an input stream filter.  The caller must delete the
-	returned object.
+	Create and return a stream filter on \p stream.  The caller must
+	delete the returned object.
 	*/
-	virtual CInputStreamFilter*
-						createInput(IInputStream*, bool adoptStream) = 0;
-
-	//! Create output filter
-	/*!
-	Create and return an output stream filter.  The caller must delete the
-	returned object.
-	*/
-	virtual COutputStreamFilter*
-						createOutput(IOutputStream*, bool adoptStream) = 0;
+	virtual IStream*	create(IStream* stream, bool adoptStream) = 0;
 };
 
 #endif
