@@ -34,3 +34,14 @@ XIOEndOfStream::getWhat() const throw()
 {
 	return format("XIOEndOfStream", "reached end of stream");
 }
+
+
+//
+// XIOWouldBlock
+//
+
+CString
+XIOWouldBlock::getWhat() const throw()
+{
+	return format("XIOWouldBlock", "stream operation would block");
+}
