@@ -24,6 +24,12 @@
 //
 #define NO_GRAB_KEYBOARD 0
 
+//
+// debugging compile flag.  when not zero the server will not
+// install low level hooks.
+//
+#define NO_LOWLEVEL_HOOKS 0
+
 //
 // extra mouse wheel stuff
 //
@@ -64,8 +70,10 @@ typedef struct tagMOUSEHOOKSTRUCTWin2000 {
 // globals
 //
 
+#if defined(_MSC_VER)
 #pragma comment(linker, "-section:shared,rws")
 #pragma data_seg("shared")
+#endif
 // all data in this shared section *must* be initialized
 
 static HINSTANCE		g_hinstance       = NULL;
@@ -76,29 +84,34 @@ static DWORD			g_threadID        = 0;
 static HHOOK			g_keyboard        = NULL;
 static HHOOK			g_mouse           = NULL;
 static HHOOK			g_getMessage      = NULL;
-static HANDLE			g_hookThreadLL    = NULL;
-static DWORD			g_hookThreadIDLL  = 0;
-static HANDLE			g_hookEventLL     = NULL;
 static HHOOK			g_keyboardLL      = NULL;
 static HHOOK			g_mouseLL         = NULL;
 static bool				g_screenSaver     = false;
-static bool				g_relay           = false;
+static EHookMode		g_mode            = kHOOK_DISABLE;
 static UInt32			g_zoneSides       = 0;
 static SInt32			g_zoneSize        = 0;
 static SInt32			g_xScreen         = 0;
 static SInt32			g_yScreen         = 0;
 static SInt32			g_wScreen         = 0;
 static SInt32			g_hScreen         = 0;
-static HCURSOR			g_cursor          = NULL;
-static DWORD			g_cursorThread    = 0;
+static WPARAM			g_deadVirtKey     = 0;
+static LPARAM			g_deadLParam      = 0;
+static WPARAM			g_oldDeadVirtKey  = 0;
+static BYTE				g_deadKeyState[256] = { 0 };
+static DWORD			g_hookThread      = 0;
+static DWORD			g_attachedThread  = 0;
 
+#if defined(_MSC_VER)
 #pragma data_seg()
+#endif
 
 // keep linker quiet about floating point stuff.  we don't use any
 // floating point operations but our includes may define some
 // (unused) floating point values.
 #ifndef _DEBUG
-extern "C" int _fltused=0;
+extern "C" {
+int _fltused=0;
+}
 #endif
 
 
@@ -108,50 +121,256 @@ extern "C" int _fltused=0;
 
 static
 void
-hideCursor(DWORD thread)
+detachThread()
 {
-	// we should be running the context of the window who's cursor
-	// we want to hide so we shouldn't have to attach thread input
-	// but we'll check to make sure.
-	g_cursorThread = thread;
-	if (g_cursorThread != 0) {
-		DWORD myThread = GetCurrentThreadId();
-		if (myThread != g_cursorThread)
-			AttachThreadInput(myThread, g_cursorThread, TRUE);
-		g_cursor = SetCursor(NULL);
-		if (myThread != g_cursorThread)
-			AttachThreadInput(myThread, g_cursorThread, FALSE);
+	if (g_attachedThread != 0 && g_hookThread != g_attachedThread) {
+		AttachThreadInput(g_hookThread, g_attachedThread, FALSE);
+		g_attachedThread = 0;
 	}
 }
 
-static
-void
-restoreCursor()
-{
-	// restore the show cursor in the window we hid it last
-	if (g_cursor != NULL && g_cursorThread != 0) {
-		DWORD myThread = GetCurrentThreadId();
-		if (myThread != g_cursorThread)
-			AttachThreadInput(myThread, g_cursorThread, TRUE);
-		SetCursor(g_cursor);
-		if (myThread != g_cursorThread)
-			AttachThreadInput(myThread, g_cursorThread, FALSE);
-	}
-	g_cursor       = NULL;
-	g_cursorThread = 0;
-}
-
 static
 bool
-keyboardHookHandler(WPARAM wParam, LPARAM lParam)
+attachThreadToForeground()
 {
+	// only attach threads if using low level hooks.  a low level hook
+	// runs in the thread that installed the hook but we have to make
+	// changes that require being attached to the target thread (which
+	// should be the foreground window).  a regular hook runs in the
+	// thread that just removed the event from its queue so we're
+	// already in the right thread.
+	if (g_hookThread != 0) {
+		HWND window    = GetForegroundWindow();
+		DWORD threadID = GetWindowThreadProcessId(window, NULL);
+		// skip if no change
+		if (g_attachedThread != threadID) {
+			// detach from previous thread
+			detachThread();
+
+			// attach to new thread
+			if (threadID != 0 && threadID != g_hookThread) {
+				AttachThreadInput(g_hookThread, threadID, TRUE);
+				g_attachedThread = threadID;
+			}
+			return true;
+		}
+	}
+	return false;
+}
+
+#if !NO_GRAB_KEYBOARD
+static
+WPARAM
+makeKeyMsg(UINT virtKey, char c)
+{
+	return MAKEWPARAM(MAKEWORD(virtKey & 0xff, (BYTE)c), 0);
+}
+
+static
+void
+keyboardGetState(BYTE keys[256])
+{
+	if (g_hookThread != 0) {
+		GetKeyboardState(keys);
+	}
+	else {
+		SHORT key;
+		for (int i = 0; i < 256; ++i) {
+			key     = GetAsyncKeyState(i);
+			keys[i] = (BYTE)((key < 0) ? 0x80u : 0);
+		}
+		key = GetKeyState(VK_CAPITAL);
+		keys[VK_CAPITAL] = (BYTE)(((key < 0) ? 0x80 : 0) | (key & 1));
+	}
+}
+
+static
+bool
+doKeyboardHookHandler(WPARAM wParam, LPARAM lParam)
+{
+	// check for dead keys.  we don't forward those to our window.
+	// instead we'll leave the key in the keyboard layout (a buffer
+	// internal to the system) for translation when the next key is
+	// pressed.  note that some systems set bit 31 to indicate a
+	// dead key and others bit 15.  nice.
+	UINT c = MapVirtualKey(wParam, 2);
+	PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | 0x00000000, c);
+	PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | (c << 8) | 0x01000000, lParam);
+	if ((c & 0x80008000u) != 0) {
+		if ((lParam & 0x80000000u) == 0) {
+			if (g_deadVirtKey == 0) {
+				// dead key press, no dead key in the buffer
+				g_deadVirtKey = wParam;
+				g_deadLParam  = lParam;
+				keyboardGetState(g_deadKeyState);
+				PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | 0x02000000, lParam);
+				return false;
+			}
+			// second dead key press in a row so let it pass
+			PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | 0x03000000, lParam);
+		}
+		else if (wParam == g_oldDeadVirtKey) {
+			// dead key release for second dead key in a row.  discard
+			// because we've already handled it.  also take it out of
+			// the keyboard buffer.
+			g_oldDeadVirtKey = 0;
+			WORD c;
+			UINT scanCode = ((lParam & 0x00ff0000u) >> 16);
+			ToAscii(wParam, scanCode, g_deadKeyState, &c, 0);
+			PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | 0x09000000, lParam);
+			return true;
+		}
+		else {
+			// dead key release
+			PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | 0x04000000, lParam);
+			return false;
+		}
+	}
+
+	// convert key to a character.  this combines a saved dead key,
+	// if any, with this key.  however, the dead key must remain in
+	// the keyboard layout for the application receiving this event
+	// so it can also convert the key to a character.  we only do
+	// this on a key press.
+	WPARAM charAndVirtKey = (wParam & 0xffu);
+	if (c != 0) {
+		// we need the keyboard state for ToAscii()
+		BYTE keys[256];
+		keyboardGetState(keys);
+
+		// ToAscii() maps ctrl+letter to the corresponding control code
+		// and ctrl+backspace to delete.  we don't want those translations
+		// so clear the control modifier state.  however, if we want to
+		// simulate AltGr (which is ctrl+alt) then we must not clear it.
+		UINT control = keys[VK_CONTROL] | keys[VK_LCONTROL] | keys[VK_RCONTROL];
+		UINT menu    = keys[VK_MENU] | keys[VK_LMENU] | keys[VK_RMENU];
+		if ((control & 0x80) == 0 || (menu & 0x80) == 0) {
+			keys[VK_LCONTROL] = 0;
+			keys[VK_RCONTROL] = 0;
+			keys[VK_CONTROL]  = 0;
+		}
+		else {
+			keys[VK_LCONTROL] = 0x80;
+			keys[VK_RCONTROL] = 0x80;
+			keys[VK_CONTROL]  = 0x80;
+			keys[VK_LMENU]    = 0x80;
+			keys[VK_RMENU]    = 0x80;
+			keys[VK_MENU]     = 0x80;
+		}
+
+		// ToAscii() needs to know if a menu is active for some reason.
+		// we don't know and there doesn't appear to be any way to find
+		// out.  so we'll just assume a menu is active if the menu key
+		// is down.
+		// FIXME -- figure out some way to check if a menu is active
+		UINT flags = 0;
+		if ((menu & 0x80) != 0)
+			flags |= 1;
+
+		// map the key event to a character.  this has the side
+		// effect of removing the dead key from the system's keyboard
+		// layout buffer.
+		WORD c        = 0;
+		UINT scanCode = ((lParam & 0x00ff0000u) >> 16);
+		int n         = ToAscii(wParam, scanCode, keys, &c, flags);
+
+		// if mapping failed and ctrl and alt are pressed then try again
+		// with both not pressed.  this handles the case where ctrl and
+		// alt are being used as individual modifiers rather than AltGr.
+		// we have to put the dead key back first, if there was one.
+		if (n == 0 && (control & 0x80) != 0 && (menu & 0x80) != 0) {
+			PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | 0x05000000, lParam);
+			if (g_deadVirtKey != 0) {
+				ToAscii(g_deadVirtKey, (g_deadLParam & 0x00ff0000u) >> 16,
+							g_deadKeyState, &c, flags);
+			}
+			keys[VK_LCONTROL] = 0;
+			keys[VK_RCONTROL] = 0;
+			keys[VK_CONTROL]  = 0;
+			keys[VK_LMENU]    = 0;
+			keys[VK_RMENU]    = 0;
+			keys[VK_MENU]     = 0;
+			n = ToAscii(wParam, scanCode, keys, &c, flags);
+		}
+
+		PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | (c << 8) | ((n & 0xff) << 16) | 0x06000000,
+							lParam);
+		switch (n) {
+		default:
+			// key is a dead key;  we're not expecting this since we
+			// bailed out above for any dead key.
+			g_deadVirtKey = wParam;
+			g_deadLParam  = lParam;
+			break;
+
+		case 0:
+			// key doesn't map to a character.  this can happen if
+			// non-character keys are pressed after a dead key.
+			break;
+
+		case 1:
+			// key maps to a character composed with dead key
+			charAndVirtKey = makeKeyMsg(wParam, (char)LOBYTE(c));
+			break;
+
+		case 2: {
+			// previous dead key not composed.  send a fake key press
+			// and release for the dead key to our window.
+			WPARAM deadCharAndVirtKey =
+							makeKeyMsg(g_deadVirtKey, (char)LOBYTE(c));
+			PostThreadMessage(g_threadID, SYNERGY_MSG_KEY,
+							deadCharAndVirtKey, g_deadLParam & 0x7fffffffu);
+			PostThreadMessage(g_threadID, SYNERGY_MSG_KEY,
+							deadCharAndVirtKey, g_deadLParam | 0x80000000u);
+
+			// use uncomposed character
+			charAndVirtKey = makeKeyMsg(wParam, (char)HIBYTE(c));
+			break;
+		}
+		}
+
+		// put back the dead key, if any, for the application to use
+		if (g_deadVirtKey != 0) {
+			ToAscii(g_deadVirtKey, (g_deadLParam & 0x00ff0000u) >> 16,
+							g_deadKeyState, &c, flags);
+			for (int i = 0; i < 256; ++i) {
+				g_deadKeyState[i] = 0;
+			}
+		}
+
+		// clear out old dead key state
+		g_deadVirtKey = 0;
+		g_deadLParam  = 0;
+	}
+
 	// forward message to our window.  do this whether or not we're
 	// forwarding events to clients because this'll keep our thread's
 	// key state table up to date.  that's important for querying
 	// the scroll lock toggle state.
-	PostThreadMessage(g_threadID, SYNERGY_MSG_KEY, wParam, lParam);
+	PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							charAndVirtKey | 0x07000000, lParam);
+	PostThreadMessage(g_threadID, SYNERGY_MSG_KEY, charAndVirtKey, lParam);
 
-	if (g_relay) {
+	// send fake key release if the user just pressed two dead keys
+	// in a row, otherwise we'll lose the release because we always
+	// return from the top of this function for all dead key releases.
+	if ((c & 0x80008000u) != 0) {
+		g_oldDeadVirtKey = wParam;
+		PostThreadMessage(g_threadID, SYNERGY_MSG_DEBUG,
+							wParam | 0x08000000, lParam);
+		PostThreadMessage(g_threadID, SYNERGY_MSG_KEY,
+							charAndVirtKey, lParam | 0x80000000u);
+	}
+
+	if (g_mode == kHOOK_RELAY_EVENTS) {
 		// let certain keys pass through
 		switch (wParam) {
 		case VK_CAPITAL:
@@ -162,8 +381,34 @@ keyboardHookHandler(WPARAM wParam, LPARAM lParam)
 			// lights may not stay synchronized.
 			break;
 
+		case VK_SHIFT:
+		case VK_LSHIFT:
+		case VK_RSHIFT:
+			// pass the shift modifiers.  if we don't do this
+			// we may not get the right dead key when caps lock
+			// is on.  for example, on the french layout (with
+			// english keycaps) on caps lock then press shift + [
+			// and q.  instead of an A with ^ above it you get an
+			// A with dots above it.
+			break;
+
+		case VK_CONTROL:
+		case VK_LCONTROL:
+		case VK_RCONTROL:
+		case VK_MENU:
+		case VK_LMENU:
+		case VK_RMENU:
+		case VK_HANGUL:
+			// pass the control and alt modifiers if using a low
+			// level hook, discard them if not.
+			if (g_hookThread == 0) {
+				return true;
+			}
+			break;
+
+
 		default:
-			// discard event
+			// discard
 			return true;
 		}
 	}
@@ -173,7 +418,16 @@ keyboardHookHandler(WPARAM wParam, LPARAM lParam)
 
 static
 bool
-mouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
+keyboardHookHandler(WPARAM wParam, LPARAM lParam)
+{
+	attachThreadToForeground();
+	return doKeyboardHookHandler(wParam, lParam);
+}
+#endif
+
+static
+bool
+doMouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
 {
 	switch (wParam) {
 	case WM_LBUTTONDOWN:
@@ -202,36 +456,50 @@ mouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
 	case WM_NCXBUTTONUP:
 		// always relay the event.  eat it if relaying.
 		PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_BUTTON, wParam, data);
-		return g_relay;
+		return (g_mode == kHOOK_RELAY_EVENTS);
 
 	case WM_MOUSEWHEEL:
-		if (g_relay) {
+		if (g_mode == kHOOK_RELAY_EVENTS) {
 			// relay event
 			PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_WHEEL, data, 0);
 		}
-		return g_relay;
+		return (g_mode == kHOOK_RELAY_EVENTS);
 
 	case WM_NCMOUSEMOVE:
 	case WM_MOUSEMOVE:
-		if (g_relay) {
-			// we want the cursor to be hidden at all times so we
-			// hide the cursor on whatever window has it.  but then
-			// we have to show the cursor whenever we leave that
-			// window (or at some later time before we stop relaying).
-			// so check the window with the cursor.  if it's not the
-			// same window that had it before then show the cursor
-			// in the last window and hide it in this window.
-			DWORD thread = GetCurrentThreadId();
-			if (thread != g_cursorThread) {
-				restoreCursor();
-				hideCursor(thread);
-			}
-
+		if (g_mode == kHOOK_RELAY_EVENTS) {
 			// relay and eat event
 			PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_MOVE, x, y);
 			return true;
 		}
-		else {
+		else if (g_mode == kHOOK_WATCH_JUMP_ZONE) {
+			// low level hooks can report bogus mouse positions that are
+			// outside of the screen.  jeez.  naturally we end up getting
+			// fake motion in the other direction to get the position back
+			// on the screen, which plays havoc with switch on double tap.
+			// CServer deals with that.  we'll clamp positions onto the
+			// screen.  also, if we discard events for positions outside
+			// of the screen then the mouse appears to get a bit jerky
+			// near the edge.  we can either accept that or pass the bogus
+			// events.  we'll try passing the events.
+			bool bogus = false;
+			if (x < g_xScreen) {
+				x     = g_xScreen;
+				bogus = true;
+			}
+			else if (x >= g_xScreen + g_wScreen) {
+				x     = g_xScreen + g_wScreen - 1;
+				bogus = true;
+			}
+			if (y < g_yScreen) {
+				y     = g_yScreen;
+				bogus = true;
+			}
+			else if (y >= g_yScreen + g_hScreen) {
+				y     = g_yScreen + g_hScreen - 1;
+				bogus = true;
+			}
+
 			// check for mouse inside jump zone
 			bool inside = false;
 			if (!inside && (g_zoneSides & kLeftMask) != 0) {
@@ -250,8 +518,8 @@ mouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
 			// relay the event
 			PostThreadMessage(g_threadID, SYNERGY_MSG_MOUSE_MOVE, x, y);
 
-			// if inside then eat the event
-			return inside;
+			// if inside and not bogus then eat the event
+			return inside && !bogus;
 		}
 	}
 
@@ -259,6 +527,15 @@ mouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
 	return false;
 }
 
+static
+bool
+mouseHookHandler(WPARAM wParam, SInt32 x, SInt32 y, SInt32 data)
+{
+	attachThreadToForeground();
+	return doMouseHookHandler(wParam, x, y, data);
+}
+
+#if !NO_GRAB_KEYBOARD
 static
 LRESULT CALLBACK
 keyboardHook(int code, WPARAM wParam, LPARAM lParam)
@@ -272,6 +549,7 @@ keyboardHook(int code, WPARAM wParam, LPARAM lParam)
 
 	return CallNextHookEx(g_keyboard, code, wParam, lParam);
 }
+#endif
 
 static
 LRESULT CALLBACK
@@ -291,15 +569,18 @@ mouseHook(int code, WPARAM wParam, LPARAM lParam)
 			// them.
 			switch (g_wheelSupport) {
 			case kWheelModern:
-				w = static_cast<SInt32>(LOWORD(info->dwExtraInfo));
+				w = static_cast<SInt16>(LOWORD(info->dwExtraInfo));
 				break;
 
 			case kWheelWin2000: {
 				const MOUSEHOOKSTRUCTWin2000* info2k =
 						(const MOUSEHOOKSTRUCTWin2000*)lParam;
-				w = static_cast<SInt32>(HIWORD(info2k->mouseData));
+				w = static_cast<SInt16>(HIWORD(info2k->mouseData));
 				break;
 			}
+
+			default:
+				break;
 			}
 		}
 
@@ -329,12 +610,13 @@ getMessageHook(int code, WPARAM wParam, LPARAM lParam)
 								SYNERGY_MSG_SCREEN_SAVER, TRUE, 0);
 			}
 		}
-		if (g_relay) {
+		if (g_mode == kHOOK_RELAY_EVENTS) {
 			MSG* msg = reinterpret_cast<MSG*>(lParam);
-			if (msg->message == g_wmMouseWheel) {
+			if (g_wheelSupport == kWheelOld && msg->message == g_wmMouseWheel) {
 				// post message to our window
 				PostThreadMessage(g_threadID,
-								SYNERGY_MSG_MOUSE_WHEEL, msg->wParam, 0);
+								SYNERGY_MSG_MOUSE_WHEEL,
+								static_cast<SInt16>(msg->wParam & 0xffffu), 0);
 
 				// zero out the delta in the message so it's (hopefully)
 				// ignored
@@ -346,7 +628,7 @@ getMessageHook(int code, WPARAM wParam, LPARAM lParam)
 	return CallNextHookEx(g_getMessage, code, wParam, lParam);
 }
 
-#if (_WIN32_WINNT >= 0x0400)
+#if (_WIN32_WINNT >= 0x0400) && defined(_MSC_VER) && !NO_LOWLEVEL_HOOKS
 
 //
 // low-level keyboard hook -- this allows us to capture and handle
@@ -354,6 +636,7 @@ getMessageHook(int code, WPARAM wParam, LPARAM lParam)
 // side, key repeats are not reported to us.
 //
 
+#if !NO_GRAB_KEYBOARD
 static
 LRESULT CALLBACK
 keyboardLLHook(int code, WPARAM wParam, LPARAM lParam)
@@ -385,6 +668,7 @@ keyboardLLHook(int code, WPARAM wParam, LPARAM lParam)
 
 	return CallNextHookEx(g_keyboardLL, code, wParam, lParam);
 }
+#endif
 
 //
 // low-level mouse hook -- this allows us to capture and handle mouse
@@ -398,9 +682,9 @@ mouseLLHook(int code, WPARAM wParam, LPARAM lParam)
 	if (code >= 0) {
 		// decode the message
 		MSLLHOOKSTRUCT* info = reinterpret_cast<MSLLHOOKSTRUCT*>(lParam);
-		SInt32 x = (SInt32)info->pt.x;
-		SInt32 y = (SInt32)info->pt.y;
-		SInt32 w = (SInt32)HIWORD(info->mouseData);
+		SInt32 x = static_cast<SInt32>(info->pt.x);
+		SInt32 y = static_cast<SInt32>(info->pt.y);
+		SInt32 w = static_cast<SInt16>(HIWORD(info->mouseData));
 
 		// handle the message
 		if (mouseHookHandler(wParam, x, y, w)) {
@@ -411,92 +695,6 @@ mouseLLHook(int code, WPARAM wParam, LPARAM lParam)
 	return CallNextHookEx(g_mouseLL, code, wParam, lParam);
 }
 
-static
-DWORD WINAPI
-getLowLevelProc(void*)
-{
-	// thread proc for low-level keyboard/mouse hooks.  this does
-	// nothing but install the hook, process events, and uninstall
-	// the hook.
-
-	// force this thread to have a message queue
-	MSG msg;
-	PeekMessage(&msg, NULL, WM_USER, WM_USER, PM_NOREMOVE);
-
-#if !NO_GRAB_KEYBOARD
-	// install low-level keyboard hook
-	g_keyboardLL = SetWindowsHookEx(WH_KEYBOARD_LL,
-								&keyboardLLHook,
-								g_hinstance,
-								0);
-	if (g_keyboardLL == NULL) {
-		// indicate failure and exit
-		g_hookThreadIDLL = 0;
-		SetEvent(g_hookEventLL);
-		return 1;
-	}
-#else
-	// keep compiler quiet
-	&keyboardLLHook;
-#endif
-
-	// install low-level mouse hook
-	g_mouseLL = SetWindowsHookEx(WH_MOUSE_LL,
-								&mouseLLHook,
-								g_hinstance,
-								0);
-	if (g_mouseLL == NULL) {
-		// indicate failure and exit
-		if (g_keyboardLL != NULL) {
-			UnhookWindowsHookEx(g_keyboardLL);
-			g_keyboardLL     = NULL;
-		}
-		g_hookThreadIDLL = 0;
-		SetEvent(g_hookEventLL);
-		return 1;
-	}
-
-	// ready
-	SetEvent(g_hookEventLL);
-
-	// message loop
-	bool done = false;
-	while (!done) {
-		switch (GetMessage(&msg, NULL, 0, 0)) {
-		case -1:
-			break;
-
-		case 0:
-			done = true;
-			break;
-
-		default:
-			TranslateMessage(&msg);
-			DispatchMessage(&msg);
-			break;
-		}
-	}
-
-	// uninstall hook
-	UnhookWindowsHookEx(g_mouseLL);
-	UnhookWindowsHookEx(g_keyboardLL);
-	g_mouseLL    = NULL;
-	g_keyboardLL = NULL;
-
-	return 0;
-}
-
-#else // (_WIN32_WINNT < 0x0400)
-
-static
-DWORD WINAPI
-getLowLevelProc(void*)
-{
-	g_hookThreadIDLL = 0;
-	SetEvent(g_hookEventLL);
-	return 1;
-}
-
 #endif
 
 static
@@ -522,6 +720,7 @@ getWheelSupport()
 	}
 
 	// not modern.  see if we've got old-style support.
+#if defined(MSH_WHEELSUPPORT)
 	UINT wheelSupportMsg    = RegisterWindowMessage(MSH_WHEELSUPPORT);
 	HWND wheelSupportWindow = FindWindow(MSH_WHEELMODULE_CLASS,
 										MSH_WHEELMODULE_TITLE);
@@ -533,6 +732,7 @@ getWheelSupport()
 			}
 		}
 	}
+#endif
 
 	// assume modern.  we don't do anything special in this case
 	// except respond to WM_MOUSEWHEEL messages.  GetSystemMetrics()
@@ -559,12 +759,8 @@ DllMain(HINSTANCE instance, DWORD reason, LPVOID)
 	}
 	else if (reason == DLL_PROCESS_DETACH) {
 		if (g_processID == GetCurrentProcessId()) {
-			if (g_keyboard   != NULL ||
-				g_mouse      != NULL ||
-				g_getMessage != NULL) {
-				uninstall();
-				uninstallScreenSaver();
-			}
+			uninstall();
+			uninstallScreenSaver();
 			g_processID = 0;
 			g_hinstance = NULL;
 		}
@@ -601,9 +797,6 @@ init(DWORD threadID)
 		g_keyboard        = NULL;
 		g_mouse           = NULL;
 		g_getMessage      = NULL;
-		g_hookThreadLL    = NULL;
-		g_hookThreadIDLL  = 0;
-		g_hookEventLL     = NULL;
 		g_keyboardLL      = NULL;
 		g_mouseLL         = NULL;
 		g_screenSaver     = false;
@@ -614,15 +807,13 @@ init(DWORD threadID)
 	g_threadID     = threadID;
 
 	// set defaults
-	g_relay        = false;
-	g_zoneSides    = 0;
-	g_zoneSize     = 0;
-	g_xScreen      = 0;
-	g_yScreen      = 0;
-	g_wScreen      = 0;
-	g_hScreen      = 0;
-	g_cursor       = NULL;
-	g_cursorThread = 0;
+	g_mode      = kHOOK_DISABLE;
+	g_zoneSides = 0;
+	g_zoneSize  = 0;
+	g_xScreen   = 0;
+	g_yScreen   = 0;
+	g_wScreen   = 0;
+	g_hScreen   = 0;
 
 	return 1;
 }
@@ -652,6 +843,10 @@ install()
 		return kHOOK_FAILED;
 	}
 
+	// discard old dead keys
+	g_deadVirtKey = 0;
+	g_deadLParam  = 0;
+
 	// check for mouse wheel support
 	g_wheelSupport = getWheelSupport();
 
@@ -663,75 +858,62 @@ install()
 								0);
 	}
 
-	// install low-level keyboard/mouse hooks, if possible.  since these
-	// hooks are called in the context of the installing thread and that
-	// thread must have a message loop but we don't want the caller's
-	// message loop to do the work, we'll fire up a separate thread
-	// just for the hooks.  note that low-level hooks are only available
-	// on windows NT SP3 and above.
-	g_hookEventLL = CreateEvent(NULL, TRUE, FALSE, NULL);
-	if (g_hookEventLL != NULL) {
-		g_hookThreadLL = CreateThread(NULL, 0, &getLowLevelProc, 0,
-								CREATE_SUSPENDED, &g_hookThreadIDLL);
-		if (g_hookThreadLL != NULL) {
-			// start the thread and wait for it to initialize
-			ResumeThread(g_hookThreadLL);
-			WaitForSingleObject(g_hookEventLL, INFINITE);
-			ResetEvent(g_hookEventLL);
-
-			// the thread clears g_hookThreadIDLL if it failed
-			if (g_hookThreadIDLL == 0) {
-				CloseHandle(g_hookThreadLL);
-				g_hookThreadLL = NULL;
-			}
-		}
-		if (g_hookThreadLL == NULL) {
-			CloseHandle(g_hookEventLL);
-			g_hookEventLL = NULL;
-		}
-	}
-
-	// install non-low-level hooks if the low-level hooks are not installed
-	if (g_hookThreadLL == NULL) {
-#if !NO_GRAB_KEYBOARD
-		g_keyboard = SetWindowsHookEx(WH_KEYBOARD,
-								&keyboardHook,
+	// install low-level hooks.  we require that they both get installed.
+#if (_WIN32_WINNT >= 0x0400) && defined(_MSC_VER) && !NO_LOWLEVEL_HOOKS
+	g_mouseLL = SetWindowsHookEx(WH_MOUSE_LL,
+								&mouseLLHook,
 								g_hinstance,
 								0);
-#else
-		// keep compiler quiet
-		&keyboardHook;
+#if !NO_GRAB_KEYBOARD
+	g_keyboardLL = SetWindowsHookEx(WH_KEYBOARD_LL,
+								&keyboardLLHook,
+								g_hinstance,
+								0);
+	if (g_mouseLL == NULL || g_keyboardLL == NULL) {
+		if (g_keyboardLL != NULL) {
+			UnhookWindowsHookEx(g_keyboardLL);
+			g_keyboardLL = NULL;
+		}
+		if (g_mouseLL != NULL) {
+			UnhookWindowsHookEx(g_mouseLL);
+			g_mouseLL = NULL;
+		}
+	}
 #endif
+#endif
+
+	// install regular hooks
+	if (g_mouseLL == NULL) {
 		g_mouse = SetWindowsHookEx(WH_MOUSE,
 								&mouseHook,
 								g_hinstance,
 								0);
 	}
-
-	// check for any failures.  uninstall all hooks on failure.
-	if (g_hookThreadLL == NULL &&
 #if !NO_GRAB_KEYBOARD
-		(g_keyboard == NULL || g_mouse == NULL)) {
-#else
-		(g_mouse == NULL)) {
+	if (g_keyboardLL == NULL) {
+		g_keyboard = SetWindowsHookEx(WH_KEYBOARD,
+								&keyboardHook,
+								g_hinstance,
+								0);
+	}
 #endif
-		if (g_keyboard != NULL) {
-			UnhookWindowsHookEx(g_keyboard);
-			g_keyboard = NULL;
-		}
-		if (g_mouse != NULL) {
-			UnhookWindowsHookEx(g_mouse);
-			g_mouse = NULL;
-		}
-		if (g_getMessage != NULL && !g_screenSaver) {
-			UnhookWindowsHookEx(g_getMessage);
-			g_getMessage = NULL;
-		}
-		g_threadID = NULL;
+
+	// check that we got all the hooks we wanted
+	if ((g_getMessage == NULL && g_wheelSupport == kWheelOld) ||
+#if !NO_GRAB_KEYBOARD
+		(g_keyboardLL == NULL && g_keyboard     == NULL) ||
+#endif
+		(g_mouseLL    == NULL && g_mouse        == NULL)) {
+		uninstall();
 		return kHOOK_FAILED;
 	}
 
-	return (g_hookThreadLL == NULL) ? kHOOK_OKAY : kHOOK_OKAY_LL;
+	if (g_keyboardLL != NULL || g_mouseLL != NULL) {
+		g_hookThread = GetCurrentThreadId();
+		return kHOOK_OKAY_LL;
+	}
+
+	return kHOOK_OKAY;
 }
 
 int
@@ -739,15 +921,21 @@ uninstall(void)
 {
 	assert(g_hinstance != NULL);
 
+	// discard old dead keys
+	g_deadVirtKey = 0;
+	g_deadLParam  = 0;
+
+	// detach from thread
+	detachThread();
+
 	// uninstall hooks
-	if (g_hookThreadLL != NULL) {
-		PostThreadMessage(g_hookThreadIDLL, WM_QUIT, 0, 0);
-		WaitForSingleObject(g_hookThreadLL, INFINITE);
-		CloseHandle(g_hookEventLL);
-		CloseHandle(g_hookThreadLL);
-		g_hookEventLL    = NULL;
-		g_hookThreadLL   = NULL;
-		g_hookThreadIDLL = 0;
+	if (g_keyboardLL != NULL) {
+		UnhookWindowsHookEx(g_keyboardLL);
+		g_keyboardLL = NULL;
+	}
+	if (g_mouseLL != NULL) {
+		UnhookWindowsHookEx(g_mouseLL);
+		g_mouseLL = NULL;
 	}
 	if (g_keyboard != NULL) {
 		UnhookWindowsHookEx(g_keyboard);
@@ -763,9 +951,6 @@ uninstall(void)
 	}
 	g_wheelSupport = kWheelNone;
 
-	// show the cursor
-	restoreCursor();
-
 	return 1;
 }
 
@@ -827,16 +1012,13 @@ setZone(SInt32 x, SInt32 y, SInt32 w, SInt32 h, SInt32 jumpZoneSize)
 }
 
 void
-setRelay(int enable)
+setMode(EHookMode mode)
 {
-	if ((enable != 0) == g_relay) {
+	if (mode == g_mode) {
 		// no change
 		return;
 	}
-	g_relay = (enable != 0);
-	if (!g_relay) {
-		restoreCursor();
-	}
+	g_mode = mode;
 }
 
 }
